from datetime import timedelta

from django_temporalio.registry import queue_workflows
from temporalio import workflow
from temporalio.common import RetryPolicy

from django_kafka.conf import settings
from django_kafka.relations_resolver.relation import SerializedRelation
from django_kafka.relations_resolver.temporalio.activities import (
    check_resolved_relations,
    resolve_relation,
)


@queue_workflows.register(settings.TEMPORAL_TASK_QUEUE)
@workflow.defn
class ResolveRelation:
    @workflow.run
    async def run(self, serialized_relation: "SerializedRelation"):
        await workflow.execute_activity(
            resolve_relation,
            serialized_relation,
            start_to_close_timeout=timedelta(minutes=10),
            retry_policy=RetryPolicy(
                maximum_attempts=0,
                backoff_coefficient=1,
            ),  # every second indefinitely
            heartbeat_timeout=timedelta(seconds=60),
        )
        await workflow.wait_condition(workflow.all_handlers_finished)


@queue_workflows.register(settings.TEMPORAL_TASK_QUEUE)
@workflow.defn
class ResolveRelations:
    @workflow.run
    async def run(self):
        await workflow.execute_activity(
            check_resolved_relations,
            start_to_close_timeout=timedelta(minutes=20),
            heartbeat_timeout=timedelta(minutes=60),
        )
        await workflow.wait_condition(workflow.all_handlers_finished)



import traceback
from typing import TYPE_CHECKING

from django.utils import timezone

from django_kafka.conf import settings
from django_kafka.consumer import Consumer, Topics
from django_kafka.dead_letter.topic import DeadLetterTopicProducer
from django_kafka.retry.header import RetryHeader
from django_kafka.retry.topic import RetryTopicConsumer

if TYPE_CHECKING:
    from confluent_kafka import cimpl

    from django_kafka.topic import TopicConsumer


class RetryTopics(Topics):
    def __init__(self, group_id: str, *topic_consumers: "TopicConsumer"):
        super().__init__(
            *(
                RetryTopicConsumer(group_id=group_id, topic_consumer=t)
                for t in topic_consumers
            ),
        )


class RetryConsumer(Consumer):
    topics: RetryTopics

    @classmethod
    def build(cls, consumer_cls: type["Consumer"]) -> type["RetryConsumer"] | None:
        """Generates RetryConsumer subclass linked to consumer class retryable topics"""
        retryable_topics = consumer_cls.topics.get_retryable(blocking=False)
        if not retryable_topics:
            return None

        group_id = consumer_cls.build_config()["group.id"]

        return type[RetryConsumer](
            f"{consumer_cls.__name__}Retry",
            (cls, consumer_cls),
            {
                "topics": RetryTopics(group_id, *retryable_topics),
                "config": {
                    **getattr(cls, "config", {}),
                    "group.id": f"{group_id}.retry",
                },
            },
        )

    @classmethod
    def build_config(cls):
        return {
            **super().build_config(),
            **settings.RETRY_CONSUMER_CONFIG,
            **getattr(cls, "config", {}),
        }

    def retry_msg(self, msg: "cimpl.Message", exc: Exception) -> (bool, bool):
        retry_topic: RetryTopicConsumer = self.get_topic(msg)
        return retry_topic.get_producer_for(msg).retry(exc), False

    def dead_letter_msg(self, msg: "cimpl.Message", exc: Exception):
        retry_topic: RetryTopicConsumer = self.get_topic(msg)
        DeadLetterTopicProducer(group_id=retry_topic.group_id, msg=msg).produce_for(
            header_summary=str(exc),
            header_detail=traceback.format_exc(),
        )

    def process_message(self, msg: "cimpl.Message"):
        retry_time = RetryHeader.get_retry_time(msg.headers())
        if retry_time and retry_time > timezone.now():
            self.pause_partition(msg, retry_time)
            return
        super().process_message(msg)



import os

from asgiref.sync import sync_to_async
from django.test import TestCase
from django.utils import timezone

from django_kafka.models import WaitingMessage
from django_kafka.relations_resolver.relation import ModelRelation
from django_kafka.tests.relations_resolver.factories import (
    OrderFactory,
    WaitingMessageFactory,
)
from django_kafka.utils.message import Message, MessageTimestamp
from example.models import Order


class WaitingMessageQuerySetTestCase(TestCase):
    def test_add_message(self):
        kafka_timestamp = int(timezone.now().timestamp() * 1000)
        kafka_msg = Message(
            topic="topic",
            key=os.urandom(16),
            value=os.urandom(16),
            headers=[("key", b"value")],
            offset=1,
            partition=1,
            timestamp=(MessageTimestamp.CREATE_TIME.value, kafka_timestamp),
        )
        relation = ModelRelation(Order, id_field="id", id_value=1)
        waiting_msg = WaitingMessage.objects.add_message(kafka_msg, relation)

        self.assertEqual(waiting_msg.topic, kafka_msg.topic())
        self.assertEqual(waiting_msg.key, kafka_msg.key())
        self.assertEqual(waiting_msg.value, kafka_msg.value())
        self.assertEqual(waiting_msg.headers, kafka_msg.headers())
        self.assertEqual(waiting_msg.offset, kafka_msg.offset())
        self.assertEqual(waiting_msg.partition, kafka_msg.partition())
        self.assertEqual(
            waiting_msg.timestamp,
            (MessageTimestamp.CREATE_TIME.value, kafka_timestamp),
        )

    def test_for_relation(self):
        relation_a = ModelRelation(Order, id_field="id", id_value=100)
        relation_b = ModelRelation(Order, id_field="id", id_value=200)

        WaitingMessageFactory.create_batch(
            5,
            relation_model_key=ModelRelation.get_model_key(Order),
            relation_id_field=relation_a.id_field,
            relation_id_value=relation_a.id_value,
            serialized_relation=relation_a.serialize(),
        )

        WaitingMessageFactory.create_batch(
            3,
            relation_model_key=ModelRelation.get_model_key(Order),
            relation_id_field=relation_b.id_field,
            relation_id_value=relation_b.id_value,
            serialized_relation=relation_b.serialize(),
        )

        self.assertEqual(
            WaitingMessage.objects.for_relation(relation_a).count(),
            5,
        )
        self.assertEqual(
            WaitingMessage.objects.for_relation(relation_b).count(),
            3,
        )

    def test_relations(self):
        WaitingMessageFactory.create_batch(
            5,
            relation_model_key=ModelRelation.get_model_key(Order),
            relation_id_field="id",
            relation_id_value=100,
            serialized_relation={},
        )

        WaitingMessageFactory.create_batch(
            3,
            relation_model_key=ModelRelation.get_model_key(Order),
            relation_id_field="id",
            relation_id_value=200,
            serialized_relation={},
        )

        self.assertEqual(
            WaitingMessage.objects.relations().count(),
            2,
        )

    def test_waiting(self):
        for status in WaitingMessage.Status:
            WaitingMessageFactory.create_batch(
                3,
                status=status.value,
                relation_model_key=ModelRelation.get_model_key(Order),
                relation_id_field="id",
                relation_id_value=100,
                serialized_relation={},
            )

        results = WaitingMessage.objects.waiting()

        for msg in results:
            self.assertEqual(msg.status, WaitingMessage.Status.WAITING)

    def test_mark_resolving(self):
        relation_a = ModelRelation(Order, id_field="id", id_value=100)
        relation_b = ModelRelation(Order, id_field="id", id_value=200)

        WaitingMessageFactory.create_batch(
            3,
            status=WaitingMessage.Status.WAITING,
            relation_model_key=ModelRelation.get_model_key(Order),
            relation_id_field=relation_a.id_field,
            relation_id_value=relation_a.id_value,
            serialized_relation={},
        )

        WaitingMessageFactory.create_batch(
            3,
            status=WaitingMessage.Status.WAITING,
            relation_model_key=ModelRelation.get_model_key(Order),
            relation_id_field=relation_b.id_field,
            relation_id_value=relation_b.id_value,
            serialized_relation={},
        )

        WaitingMessage.objects.mark_resolving(relation_a)

        for msg in WaitingMessage.objects.for_relation(relation_a):
            self.assertEqual(msg.status, WaitingMessage.Status.RESOLVING)

        for msg in WaitingMessage.objects.for_relation(relation_b):
            self.assertEqual(msg.status, WaitingMessage.Status.WAITING)

    async def test_aiter_relations_to_resolve(self):
        for status in WaitingMessage.Status:
            await sync_to_async(WaitingMessageFactory.create_batch)(
                3,
                status=status.value,
                relation_model_key=ModelRelation.get_model_key(Order),
                relation_id_field="id",
                relation_id_value=1000,
                serialized_relation={},
            )

        relation = await sync_to_async(OrderFactory.create)(id=1)
        for status in WaitingMessage.Status:
            await sync_to_async(WaitingMessageFactory.create_batch)(
                3,
                status=status.value,
                relation_model_key=ModelRelation.get_model_key(Order),
                relation_id_field="id",
                relation_id_value=relation.id,
                serialized_relation={},
            )

        results = [
            msg async for msg in WaitingMessage.objects.aiter_relations_to_resolve()
        ]

        self.assertEqual(len(results), 2)  # 1 existing and 1 non-existing



from datetime import datetime
from typing import Optional

from django.utils import timezone

from django_kafka.utils.message import Header


class RetryHeader(Header):
    MESSAGE = "RETRY_MESSAGE"
    TIMESTAMP = "RETRY_TIMESTAMP"

    @classmethod
    def get_retry_time(cls, headers: list[tuple] | None) -> Optional["datetime"]:
        """returns the retry time from the message headers"""
        header = RetryHeader.get(headers, RetryHeader.TIMESTAMP)
        try:
            epoch = float(header)
        except (TypeError, ValueError):
            return None
        else:
            return datetime.fromtimestamp(epoch, tz=timezone.get_current_timezone())



from unittest.mock import Mock

from django.core.management import CommandError
from django.test import SimpleTestCase

from django_kafka.management.commands.errors import substitute_error


class SubstituteErrorTestCase(SimpleTestCase):
    def test_substitute(self):
        class CustomError(Exception):
            pass

        errors = {
            ValueError: "value error",
            KeyError: "key error",
            CustomError: "custom error",
        }
        decorator = substitute_error(errors.keys(), CommandError)

        for error, msg in errors.items():
            func = Mock(side_effect=error(msg))

            with self.assertRaisesMessage(CommandError, msg):
                decorator(func)()



from unittest.mock import AsyncMock, Mock, PropertyMock, patch

from django.test import SimpleTestCase

from django_kafka.relations_resolver.processor.base import MessageProcessor
from django_kafka.relations_resolver.relation import Relation


@patch.multiple(
    "django_kafka.relations_resolver.relation.Relation",
    __abstractmethods__=set(),
)
@patch(
    "django_kafka.kafka.relations_resolver.processor",
    new_callable=PropertyMock(
        return_value=AsyncMock(spec=MessageProcessor, __abstractmethods__=set()),
    ),
)
class RelationTestCase(SimpleTestCase):
    async def test_amark_resolving_uses_msg_processor(self, mock_msg_processor):
        relation = Relation()
        await relation.amark_resolving()
        mock_msg_processor.amark_resolving.assert_called_once_with(relation)

    @patch(
        "django_kafka.relations_resolver.relation.sync_to_async",
        return_value=AsyncMock(),
    )
    async def test_aidentifier_converts_identifier_method(
        self,
        mock_sync_to_async,
        mock_msg_processor,
    ):
        relation = Relation()
        await relation.aidentifier()
        mock_sync_to_async.assert_called_once_with(relation.identifier)

    @patch("django_kafka.relations_resolver.relation.RelationType")
    def test_type_returns_relation_type_name(
        self,
        mock_relation_type,
        mock_msg_processor,
    ):
        self.assertEqual(Relation().type(), mock_relation_type.return_value.name)

    async def test_ahas_waiting_messages_uses_msg_processor(self, mock_msg_processor):
        relation = Relation()
        await relation.ahas_waiting_messages()
        mock_msg_processor.aexists.assert_called_once_with(relation)

    async def test_aadd_message_uses_msg_processor(self, mock_msg_processor):
        relation = Relation()
        msg = Mock()
        await relation.aadd_message(msg)
        mock_msg_processor.aadd_message.assert_called_once_with(msg, relation)

    async def test_resolve_uses_msg_processor(self, mock_msg_processor):
        relation = Relation()
        await relation.aresolve()
        mock_msg_processor.aprocess_messages.assert_called_once_with(relation)



import datetime

from django.test import SimpleTestCase
from django.utils import timezone

from django_kafka.consumer.managers import PauseManager, RetryManager
from django_kafka.tests.utils import message_mock


class PauseManagerTestCase(SimpleTestCase):
    def test_get_msg_partition(self):
        mock_msg = message_mock()
        manager = PauseManager()

        tp = manager.get_msg_partition(mock_msg)

        self.assertEqual(tp.topic, mock_msg.topic())
        self.assertEqual(tp.partition, mock_msg.partition())

    def test_set(self):
        mock_msg = message_mock(offset=1000)
        manager = PauseManager()

        tp = manager.set(mock_msg, timezone.now())

        result_tp = manager.get_msg_partition(mock_msg)
        self.assertEqual(result_tp, tp)
        self.assertEqual(result_tp.offset, 1000)

    def test_pop_ready(self):
        manager = PauseManager()
        mock_msg_1 = message_mock(partition=1, offset=1000)
        mock_msg_2 = message_mock(partition=2, offset=2000)

        manager.set(mock_msg_1, timezone.now() - datetime.timedelta(minutes=1))
        manager.set(mock_msg_2, timezone.now() + datetime.timedelta(minutes=1))

        ready_tps = list(manager.pop_ready())
        self.assertEqual(len(ready_tps), 1)
        self.assertEqual(ready_tps[0], manager.get_msg_partition(mock_msg_1))
        self.assertEqual(ready_tps[0].offset, 1000)
        self.assertEqual(list(manager.pop_ready()), [])  # empty the second time


class RetryManagerTestCase(SimpleTestCase):
    def test_get_msg_partition(self):
        mock_msg = message_mock()
        manager = RetryManager()

        tp = manager.get_msg_partition(mock_msg)

        self.assertEqual(tp.topic, mock_msg.topic())
        self.assertEqual(tp.partition, mock_msg.partition())
        self.assertEqual(tp.offset, mock_msg.offset())

    def test_next(self):
        mock_msg = message_mock()
        manager = RetryManager()

        self.assertEqual(manager.next(mock_msg), 1)
        self.assertEqual(manager.next(mock_msg), 2)
        self.assertEqual(manager.next(mock_msg), 3)

    def test_next__resets_for_new_offset(self):
        """tests retry attempt resets if requested for a new offset"""
        mock_msg = message_mock(offset=0)
        manager = RetryManager()

        self.assertEqual(manager.next(mock_msg), 1)

        mock_msg.offset.return_value = 1

        self.assertEqual(manager.next(mock_msg), 1)
        self.assertEqual(manager.next(mock_msg), 2)



from typing import Any


class DjangoKafkaError(Exception):
    def __init__(self, *args, context: Any | None = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.context = context


class TopicNotRegisteredError(DjangoKafkaError):
    pass



import datetime
from unittest import mock

from django.test import TestCase

from django_kafka.models import KeyOffsetTracker
from django_kafka.retry.settings import RetrySettings
from django_kafka.tests.utils import message_mock
from django_kafka.topic import Topic


class RetrySettingTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.mock_msg = message_mock()
        super().setUpClass()

    def test_should_retry__include(self):
        settings = RetrySettings(max_retries=5, delay=60, include=[ValueError])

        self.assertEqual(
            settings.should_retry(self.mock_msg, attempt=0, exc=ValueError()),
            True,
        )
        self.assertEqual(
            settings.should_retry(self.mock_msg, attempt=0, exc=IndexError()),
            False,
        )

    def test_should_retry__exclude(self):
        settings = RetrySettings(max_retries=5, delay=60, exclude=[ValueError])

        self.assertEqual(
            settings.should_retry(self.mock_msg, attempt=0, exc=ValueError()),
            False,
        )
        self.assertEqual(
            settings.should_retry(self.mock_msg, attempt=0, exc=IndexError()),
            True,
        )

    def test_should_retry__no_include_exclude(self):
        settings = RetrySettings(max_retries=5, delay=60)

        self.assertEqual(
            settings.should_retry(self.mock_msg, attempt=0, exc=ValueError()),
            True,
        )

    def test_attempts_exceeded(self):
        settings = RetrySettings(max_retries=5, delay=60)

        self.assertEqual(settings.attempts_exceeded(6), True)
        self.assertEqual(settings.attempts_exceeded(5), False)
        self.assertEqual(settings.attempts_exceeded(4), False)

    @mock.patch("django.utils.timezone.now")
    def test_get_retry_time__fixed_delay(self, mock_now):
        settings = RetrySettings(max_retries=5, delay=60, backoff=False)
        now = datetime.datetime.fromtimestamp(1000, datetime.UTC)
        mock_now.return_value = now

        self.assertEqual(
            settings.get_retry_time(1),
            now + datetime.timedelta(seconds=60),
        )
        self.assertEqual(
            settings.get_retry_time(2),
            now + datetime.timedelta(seconds=60),
        )
        self.assertEqual(
            settings.get_retry_time(3),
            now + datetime.timedelta(seconds=60),
        )
        self.assertEqual(
            settings.get_retry_time(4),
            now + datetime.timedelta(seconds=60),
        )

    @mock.patch("django.utils.timezone.now")
    def test_get_retry_time__backoff_delay(self, mock_now):
        settings = RetrySettings(max_retries=5, delay=60, backoff=True)
        now = datetime.datetime.fromtimestamp(1000, datetime.UTC)
        mock_now.return_value = now

        self.assertEqual(
            settings.get_retry_time(1),
            now + datetime.timedelta(seconds=60),
        )
        self.assertEqual(
            settings.get_retry_time(2),
            now + datetime.timedelta(seconds=120),
        )
        self.assertEqual(
            settings.get_retry_time(3),
            now + datetime.timedelta(seconds=240),
        )
        self.assertEqual(
            settings.get_retry_time(4),
            now + datetime.timedelta(seconds=480),
        )

    def test_should_log(self):
        settings = RetrySettings(max_retries=5, delay=60, log_every=5)

        self.assertEqual(settings.should_log(1), False)
        self.assertEqual(settings.should_log(2), False)
        self.assertEqual(settings.should_log(5), True)
        self.assertEqual(settings.should_log(10), True)

    def test_skip_by_offset(self):
        msg = message_mock()
        model_instance = KeyOffsetTracker.objects.create(
            topic=msg.topic(),
            key=msg.key(),
            offset=msg.offset(),
            timestamp=msg.timestamp(),
        )

        # not configured to use offset
        settings = RetrySettings(
            max_retries=5,
            delay=60,
            backoff=False,
            use_offset_tracker=False,
        )
        for error in settings.relational_errors:
            self.assertFalse(settings.skip_by_offset(msg, error()))

        # configured to use offset, but exception is not among accepted
        settings = RetrySettings(
            max_retries=5,
            delay=60,
            backoff=False,
            use_offset_tracker=True,
        )
        self.assertFalse(settings.skip_by_offset(msg, Exception()))

        # configured to use offset, exceptions are valid but no future offset
        settings = RetrySettings(
            max_retries=5,
            delay=60,
            backoff=False,
            use_offset_tracker=True,
        )
        for error in settings.relational_errors:
            self.assertFalse(settings.skip_by_offset(self.mock_msg, error()))

        # configured to use offset and everything is aligned
        settings = RetrySettings(
            max_retries=5,
            delay=60,
            backoff=False,
            use_offset_tracker=True,
        )

        model_instance.offset += 1
        model_instance.save(update_fields=["offset"])

        for error in settings.relational_errors:
            self.assertTrue(settings.skip_by_offset(msg, error()))

    @mock.patch(
        "django_kafka.retry.settings.RetrySettings.skip_by_offset",
        return_value=True,
    )
    def test_should_retry_calls_skip_by_offset(self, skip_by_offset):
        settings = RetrySettings(
            max_retries=5,
            delay=60,
            backoff=False,
            use_offset_tracker=True,
        )
        error = Exception()
        result = settings.should_retry(self.mock_msg, 1, error)

        skip_by_offset.assert_called_once_with(self.mock_msg, error)
        self.assertIs(result, False)


class RetryDecoratorTestCase(TestCase):
    def test_retry_decorator(self):
        class TopicA(Topic):
            pass

        retry = RetrySettings(max_retries=5, delay=60)
        result = retry(TopicA)

        self.assertIs(result, TopicA)
        self.assertIsNotNone(TopicA.retry_settings)
        self.assertEqual(
            [TopicA.retry_settings.max_retries, TopicA.retry_settings.delay],
            [5, 60],
        )



from collections.abc import Iterable
from typing import TYPE_CHECKING, ClassVar

from django_kafka import kafka
from django_kafka.consumer import Consumer, Topics
from django_kafka.exceptions import DjangoKafkaError
from django_kafka.topic import TopicConsumer

if TYPE_CHECKING:
    from confluent_kafka import cimpl

    from django_kafka.topic import Topic


class KeyOffsetTrackerTopic(TopicConsumer):
    name = None

    def __init__(self, topic_name: str):
        self.name = topic_name
        super().__init__()

    def consume(self, msg: "cimpl.Message"):
        from django_kafka.models import KeyOffsetTracker

        KeyOffsetTracker.objects.log_msg_offset(msg)
        return True


class KeyOffsetTrackerConsumer(Consumer):
    topics = None
    config: ClassVar = {
        "auto.offset.reset": "earliest",
        "enable.auto.offset.store": False,
    }

    def start(self):
        self.run_sanity_checks()
        super().start()

    @classmethod
    def activate(cls, group_id: str) -> None:
        cls.setup(group_id)
        cls.run_sanity_checks()
        kafka.consumers()(cls)

    @classmethod
    def setup(cls, group_id: str):
        cls.config = {**cls.config, "group.id": group_id}
        cls.topics = Topics(*cls.tracked_topics())

    @classmethod
    def run_sanity_checks(cls):
        if not cls.build_config().get("group.id"):
            raise DjangoKafkaError("'group.id' must be specified.")

        if not cls.topics:
            raise DjangoKafkaError(
                "Key offset tracker consumer is registered but there are no topics "
                "configured to use it.",
            )

    @classmethod
    def tracked_topics(cls) -> Iterable["Topic"]:
        topics = []
        for key in kafka.consumers:
            for topic in kafka.consumers[key].topics:
                if topic.retry_settings and topic.retry_settings.use_offset_tracker:
                    topics.append(topic.name)
        return map(KeyOffsetTrackerTopic, set(topics))



import logging
from collections.abc import Iterator
from enum import IntEnum
from typing import TYPE_CHECKING

from asgiref.sync import async_to_sync

from django_kafka.relations_resolver.processor import get_message_processor

if TYPE_CHECKING:
    from confluent_kafka import cimpl

    from django_kafka.relations_resolver.processor.base import MessageProcessor
    from django_kafka.relations_resolver.relation import Relation

logger = logging.getLogger(__name__)


class RelationResolver:
    processor: "MessageProcessor"

    class Action(IntEnum):
        CONTINUE = 1
        SKIP = 2
        PAUSE = 3

    def __init__(self):
        self.processor = get_message_processor()

    async def await_for_relation(self, msg: "cimpl.Message", relation: "Relation"):
        """
        Send the message to the comfortable place to wait for a relation.
        """
        await relation.aadd_message(msg)

    async def aresolve(
        self,
        relations: Iterator["Relation"],
        msg: "cimpl.Message",
    ) -> Action:
        logger.debug("Check for missing relations.")
        for relation in relations:
            if not await relation.aexists():
                logger.debug("Relation is missing - send message to wait.")
                await self.await_for_relation(msg, relation)
                return self.Action.SKIP

            if await relation.ahas_waiting_messages():
                logger.debug("Relation exists but has waiting messages.")
                return self.Action.PAUSE

        logger.debug("No relations missing.")
        return self.Action.CONTINUE

    def resolve(self, relations: Iterator["Relation"], msg: "cimpl.Message") -> Action:
        return async_to_sync(self.aresolve)(relations, msg)

    async def aresolve_relation(self, relation: "Relation"):
        await relation.aresolve()



from django.db import models
from django.utils.translation import gettext_lazy as _


class KafkaConnectSkipQueryset(models.QuerySet):
    def update(self, **kwargs) -> int:
        kwargs.setdefault("kafka_skip", False)
        return super().update(**kwargs)


class KafkaConnectSkipModel(models.Model):
    """
    For models (tables) which have Kafka Connect source connectors attached and require
    a flag to suppress message production.

    The Kafka Connect connector should filter out events based on the kafka_skip flag
    provided in this model.

    Any update to the model instance will reset the kafka_skip flag to False, if not
    explicitly set.

    This flag can help overcome infinite event loops during bidirectional data sync when
    using Kafka. See README.md for more information.
    """

    kafka_skip = models.BooleanField(
        _("Kafka skip"),
        help_text=_(
            "Used by Kafka Connect to suppress event creation."
            "\nGets reset to False on .save() method call, unless explicitly set.",
        ),
        default=False,
    )
    objects = KafkaConnectSkipQueryset.as_manager()

    class Meta:
        abstract = True
        base_manager_name = "objects"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._reset_kafka_skip = False  # set True for DB fetched instances in from_db

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if key == "kafka_skip":
            self._reset_kafka_skip = False

    def save(
        self,
        force_insert=False,
        force_update=False,
        using=None,
        update_fields=None,
    ):
        if self._reset_kafka_skip:
            self.kafka_skip = False
            if update_fields and "kafka_skip" not in update_fields:
                update_fields = ["kafka_skip", *update_fields]

        super().save(
            force_insert=force_insert,
            force_update=force_update,
            using=using,
            update_fields=update_fields,
        )
        self._reset_kafka_skip = True

    @classmethod
    def from_db(cls, *args, **kwargs):
        """
        Used by Django's QuerySet to initialize instances fetched from db, and won't
        be called by QuerySet.create or direct initialization.

        The kafka_skip value stored in the DB should not be reused; instead it should be
        reset to False before updating - unless explicitly set.
        """
        instance = super().from_db(*args, **kwargs)
        instance._reset_kafka_skip = True
        return instance

    def refresh_from_db(self, *args, **kwargs):
        super().refresh_from_db(*args, **kwargs)
        self._reset_kafka_skip = True



from pydoc import locate
from typing import TYPE_CHECKING

from django_kafka.conf import settings
from django_kafka.exceptions import DjangoKafkaError

if TYPE_CHECKING:
    from django_kafka.relations_resolver.processor.base import MessageProcessor


def get_message_processor() -> "MessageProcessor":
    if not (processor_cls := locate(settings.RELATION_RESOLVER_PROCESSOR)):
        raise DjangoKafkaError(f"{settings.RELATION_RESOLVER_PROCESSOR} not found.")
    return processor_cls()



from unittest.mock import Mock

from django.test import SimpleTestCase

from django_kafka.consumer import Topics
from django_kafka.retry.settings import RetrySettings


class TopicsTestCase(SimpleTestCase):
    def test_get_retryable(self):
        blocking_tc = Mock(
            retry_settings=RetrySettings(delay=1, max_retries=1, blocking=True),
        )
        non_blocking_tc = Mock(
            retry_settings=RetrySettings(delay=1, max_retries=1, blocking=False),
        )

        topics = Topics(blocking_tc, non_blocking_tc)

        self.assertEqual(topics.get_retryable(blocking=True), [blocking_tc])
        self.assertEqual(topics.get_retryable(blocking=False), [non_blocking_tc])



from abc import ABC, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, TypedDict

from asgiref.sync import sync_to_async
from django.apps import apps
from temporalio import workflow

from django_kafka import kafka

if TYPE_CHECKING:
    from confluent_kafka import cimpl

with workflow.unsafe.imports_passed_through():
    from django.db.models import Model


class Relation(ABC):
    @abstractmethod
    def identifier(self) -> str: ...

    @abstractmethod
    async def aexists(self) -> bool: ...

    @abstractmethod
    def serialize(self) -> dict: ...

    @abstractmethod
    def deserialize(self, **kwargs) -> "Relation": ...

    async def amark_resolving(self):
        await kafka.relations_resolver.processor.amark_resolving(self)

    async def aidentifier(self) -> str:
        return await sync_to_async(self.identifier)()

    def type(self) -> str:
        return RelationType(self.__class__).name

    async def ahas_waiting_messages(self) -> bool:
        return await kafka.relations_resolver.processor.aexists(self)

    async def aadd_message(self, msg: "cimpl.Message"):
        await kafka.relations_resolver.processor.aadd_message(msg, self)

    async def aresolve(self):
        await kafka.relations_resolver.processor.aprocess_messages(self)


class ModelRelation(Relation):
    model: type[Model]
    id_value: int
    id_field: str
    model_key: str

    def __init__(self, model: type[Model], id_field: str, id_value: int | str):
        self.model = model
        self.id_field = id_field
        self.id_value = id_value
        self.model_key = self.get_model_key(model)

    def identifier(self):
        return f"{self.model_key}-{self.id_value}".lower()

    async def aexists(self) -> bool:
        return await self.model.objects.filter(
            **{self.id_field: self.id_value},
        ).aexists()

    def serialize(self) -> "SerializedRelation":
        return {
            "type": self.type(),
            "kwargs": {
                "model": self.model_key,
                "id_value": self.id_value,
                "id_field": self.id_field,
            },
        }

    @classmethod
    def deserialize(
        cls,
        model: str,
        id_field: str,
        id_value: int | str,
    ) -> "ModelRelation":
        return cls(apps.get_model(model), id_field, id_value)

    @staticmethod
    def get_model_key(model: type[Model]) -> str:
        return f"{model._meta.app_label}.{model._meta.model_name}"


class RelationType(Enum):
    MODEL = ModelRelation

    @classmethod
    def instance(cls, serialized_relation: "SerializedRelation") -> Relation:
        return cls[serialized_relation["type"]].value.deserialize(
            **serialized_relation["kwargs"],
        )


class SerializedRelation(TypedDict):
    type: str
    kwargs: dict



from collections.abc import AsyncIterator
from typing import TYPE_CHECKING, ClassVar

from django.db import models
from django.utils.translation import gettext_lazy as _

from django_kafka.relations_resolver.relation import RelationType
from django_kafka.utils.message import MessageTimestamp

if TYPE_CHECKING:
    from confluent_kafka import cimpl

    from django_kafka.relations_resolver.relation import ModelRelation


class KeyOffsetTrackerQuerySet(models.QuerySet):
    def log_msg_offset(self, msg) -> tuple["KeyOffsetTracker", bool]:
        return self.update_or_create(
            topic=msg.topic(),
            key=msg.key(),
            defaults={
                "offset": msg.offset(),
                "timestamp": msg.timestamp(),
            },
        )

    def has_future_offset(self, msg) -> bool:
        return self.filter(
            topic=msg.topic(),
            key=msg.key(),
            offset__gt=msg.offset(),
        ).exists()


class KeyOffsetTracker(models.Model):
    topic = models.CharField(_("topic"), max_length=256)
    key = models.BinaryField(_("key"))
    offset = models.PositiveIntegerField(_("offset"))
    create_time = models.DateTimeField(_("create time"), null=True)
    log_append_time = models.DateTimeField(_("log append time"), null=True)

    objects = KeyOffsetTrackerQuerySet.as_manager()

    class Meta:
        verbose_name = _("key offset tracker")
        verbose_name_plural = _("key offsets tracker")
        indexes: ClassVar = [
            models.Index(
                fields=["topic", "key"],
                include=["offset"],
                name="%(app_label)s_key_offset_index",
            ),
        ]

    def __str__(self):
        return f"{self.topic}:{self.key}:{self.offset}"

    @property
    def timestamp(self):
        if self.log_append_time:
            return self.log_append_time
        return self.create_time or self.log_append_time

    @timestamp.setter
    def timestamp(self, value: tuple[MessageTimestamp, int]):
        if not value:
            return

        timestamp_type = value[0]

        if timestamp_type == MessageTimestamp.CREATE_TIME:
            self.create_time = MessageTimestamp.to_datetime(value)
            self.log_append_time = None

        elif timestamp_type == MessageTimestamp.LOG_APPEND_TIME:
            self.create_time = None
            self.log_append_time = MessageTimestamp.to_datetime(value)

        else:
            self.create_time = None
            self.log_append_time = None


class WaitingMessageQuerySet(models.QuerySet):
    def add_message(self, msg: "cimpl.Message", relation: "ModelRelation"):
        from django_kafka.relations_resolver.relation import ModelRelation

        return self.create(
            key=msg.key(),
            value=msg.value(),
            timestamp=msg.timestamp(),
            topic=msg.topic(),
            partition=msg.partition(),
            headers=msg.headers(),
            offset=msg.offset(),
            relation_model_key=ModelRelation.get_model_key(relation.model),
            relation_id_field=relation.id_field,
            relation_id_value=relation.id_value,
            serialized_relation=relation.serialize(),
        )

    def for_relation(self, relation):
        from django_kafka.relations_resolver.relation import ModelRelation

        return self.filter(
            relation_model_key=ModelRelation.get_model_key(relation.model),
            relation_id_field=relation.id_field,
            relation_id_value=relation.id_value,
        ).order_by("topic", "partition", "offset")

    def relations(self):
        return (
            self.order_by(
                "relation_model_key",
                "relation_id_field",
                "relation_id_value",
            )
            .distinct("relation_model_key", "relation_id_field", "relation_id_value")
            .only("relation_model_key", "relation_id_field", "relation_id_value")
        )

    def waiting(self):
        return self.filter(status=self.model.Status.WAITING)

    def mark_resolving(self, relation):
        self.for_relation(relation).update(status=self.model.Status.RESOLVING)

    async def aiter_relations_to_resolve(
        self,
        chunk_size=500,
    ) -> AsyncIterator["WaitingMessage"]:
        async for relation in (
            self.waiting().relations().aiterator(chunk_size=chunk_size)
        ):
            yield relation


class WaitingMessage(models.Model):
    class Status(models.IntegerChoices):
        WAITING = 1
        RESOLVING = 2
        RESOLVED = 3

    status = models.IntegerField(
        _("status"),
        choices=Status.choices,
        default=Status.WAITING,
    )

    key = models.BinaryField(_("key"))
    value = models.BinaryField(_("value"))
    timestamp = models.JSONField(_("kafka timestamp"))
    topic = models.TextField(_("topic"))
    partition = models.TextField(_("partition"))
    offset = models.TextField(_("offset"))
    _headers = models.JSONField(_("headers"), null=True)

    relation_model_key = models.CharField(_("relation model key"), max_length=255)
    relation_id_field = models.CharField(_("relation id field"), max_length=256)
    relation_id_value = models.CharField(_("relation id value"), max_length=256)

    serialized_relation = models.JSONField(_("relation kwargs"))

    objects = WaitingMessageQuerySet.as_manager()

    class Meta:
        verbose_name = _("waiting message")
        verbose_name_plural = _("waiting messages")
        ordering = ("offset",)

    def __str__(self):
        return (
            f"{self.relation_model_key}"
            f"({self.relation_id_field}={self.relation_id_value})"
        )

    @property
    def headers(self) -> list[tuple[str, bytes]] | None:
        if not self._headers:
            return None
        return [(k, bytes(v) if isinstance(v, list) else v) for k, v in self._headers]

    @headers.setter
    def headers(self, value: list[tuple[str, bytes]]):
        if not value:
            self._headers = None
        else:
            self._headers = [
                (k, list(v) if isinstance(v, bytes) else v) for k, v in value
            ]

    def relation(self):
        return RelationType.instance(self.serialized_relation)



from collections.abc import Callable, Iterable
from functools import wraps


def substitute_error(
    errors: Iterable[type[Exception]],
    substitution: type[Exception],
) -> Callable:
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except tuple(errors) as original_error:
                raise substitution(original_error) from original_error

        return wrapper

    return decorator



from typing import TYPE_CHECKING

from django_temporalio.client import init_client
from temporalio.common import RetryPolicy

from django_kafka.conf import settings
from django_kafka.relations_resolver.daemon import RelationResolverDaemon
from django_kafka.relations_resolver.temporalio.workflows import ResolveRelation

if TYPE_CHECKING:
    from django_kafka.relations_resolver.relation import Relation


class TemporalDaemon(RelationResolverDaemon):
    task_queue = settings.TEMPORAL_TASK_QUEUE

    async def aresolve_relation(self, relation: "Relation"):
        await (await init_client()).start_workflow(
            ResolveRelation.run,
            relation.serialize(),
            id=await relation.aidentifier(),
            task_queue=self.task_queue,
            retry_policy=RetryPolicy(maximum_attempts=0),  # indefinitely
        )



from django.test import SimpleTestCase
from django.utils import timezone

from django_kafka.retry.header import RetryHeader


class RetryHeaderTestCase(SimpleTestCase):
    def test_get_retry_time(self):
        now = timezone.now()
        headers = [(RetryHeader.TIMESTAMP, str(now.timestamp()))]

        self.assertEqual(RetryHeader.get_retry_time(headers), now)

    def test_get_retry_time__none_when_unavailable(self):
        headers1 = [(RetryHeader.TIMESTAMP, "abc")]
        headers2 = []
        headers3 = None

        self.assertEqual(RetryHeader.get_retry_time(headers1), None)
        self.assertEqual(RetryHeader.get_retry_time(headers2), None)
        self.assertEqual(RetryHeader.get_retry_time(headers3), None)
